// Copyright (C) 2008 Helmut Hirtes
// This file is a free extension of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#include "XAnimatorCameraSceneNode.h"


// DEBUG
//#define HH_DEBUG
#ifdef  HH_DEBUG
#include <iostream>
#define HH_COUT(_PARAM) std::cout << __func__ << " ==> " << _PARAM << std::endl;
//#define HH_ENTER HH_COUT("enter")
#define HH_ENTER
#define HH_VECTOR(_TEXT,_V) HH_COUT( _TEXT " = [ " << _V.X << " | " << _V.Y << " | " << _V.Z << " ]" )
#else
#define HH_COUT(_PARAM)
#define HH_ENTER
#define HH_VECTOR(_TEXT,_V)
#endif


namespace irr
{
namespace scene
{


//! Reads attributes of the scene node animator.
void XNodePositionData::deserializeAttributes
                        ( io::IAttributes *in,
                          io::SAttributeReadWriteOptions *options )
{ HH_ENTER
  _Backward    = in->getAttributeAsFloat              (   "_Backward" );
  ev_Backward  = in->getAttributeAsVector3d           ( "ev_Backward" );
  v_Fixpoint   = in->getAttributeAsVector3d           (  "v_Fixpoint" );
  _Forward     = in->getAttributeAsFloat              (   "_Forward" );
  ev_Forward   = in->getAttributeAsVector3d           ( "ev_Forward" );
  setViewpoint();  // may be no good idea to load a reference here ...
  _Orientation = (eT_Orientation)in->getAttributeAsInt(   "_Orientation" );
  v_Position   = in->getAttributeAsVector3d           (  "v_Position" );
  _Speed       = in->getAttributeAsFloat              (   "_Speed" );
}


//! Writes attributes of the scene node animator.
void XNodePositionData::serializeAttributes
                        ( io::IAttributes *out,
                          io::SAttributeReadWriteOptions *options ) const
{ HH_ENTER
  out->addFloat   (   "_Backward",         _Backward );
  out->addVector3d( "ev_Backward",       ev_Backward );
  out->addVector3d(  "v_Fixpoint",        v_Fixpoint );
  out->addFloat   (   "_Forward",          _Forward );
  out->addVector3d( "ev_Forward",        ev_Forward );
  // may be no good idea to store a reference here ... ISceneNode *p_Viewpoint;
  out->addInt     (   "_Orientation", (u32)_Orientation );  // eT_Orientation
  out->addVector3d(  "v_Position",        v_Position );
  out->addFloat   (   "_Speed",            _Speed );
}


//! Constructor.
//! \param forward_direction The forward direction.
//! \param backward_direction The backward direction.
XNodePosition::XNodePosition( s32 id )
:IReferenceCounted()
,_Id( id )
,_InitIsDone( false )
,_Length( 0.0 )
{ HH_ENTER
  // empty
}


//! Reads attributes of the scene node animator.
void XNodePosition::deserializeAttributes
                    ( io::IAttributes *in,
                      io::SAttributeReadWriteOptions *options )
{ HH_ENTER
  _Id = in->getAttributeAsInt( "_Id" );
  _Data.deserializeAttributes( in, options );
  // prepare for init.
  _InitIsDone = false;
  _Length = 0.0;
}


//! Calculates the position between the last node and this node.
//! \param node The data of the previous node.
//! \param position Is set to the calcualted position.
//! \param t The relative position (range from 0.0 to 1.0).
void XNodePosition::getPosition( const XNodePositionData &node,
                                 core::vector3df &position,
                                 f32 t ) const
{ HH_ENTER
  // prepare
  const core::vector3df &P0 =  node.v_Position;
  const core::vector3df &P3 = _Data.v_Position;
  const core::vector3df  P1 = P0 +  node._Forward  *  node.ev_Forward;
  const core::vector3df  P2 = P3 + _Data._Backward * _Data.ev_Backward;
  // calculate the position
  if ( t <= 0 )
    position = P0;
  else
  if ( t >= 1 )
    position = P3;
  else
  {
    const f32 _t   = t;
    const f32 _t2  = _t   * _t;
    const f32 _t3  = _t2  * _t;
    const f32 _1t  = 1.0f - _t;
    const f32 _1t2 = _1t  * _1t;
    const f32 _1t3 = _1t2 * _1t;
    position =          _1t3 * P0 +
        (_t +_t +_t)  * _1t2 * P1 +
        (_t2+_t2+_t2) * _1t  * P2 +
                 _t3  *        P3;
  }
}


//! Calculates the position between the last node and this node.
//! Note that the vector of direction will be zero if both node
//! position are equal.
//! Note that the vector of centrifugal force may be a zero vector in
//! the case that the actual position has no radius.
//! \param node The data of the previous node.
//! \param position Is set to the calcualted position.
//! \param direction The vector of direction.
//! \param centrifugal_force The vector of centrifugal force.
//! \param t The relative position (range from 0.0 to 1.0).
//! \return true if the vector of centrifugal force is not a zero vector.
bool XNodePosition::getPosition( const XNodePositionData &node,
                                 core::vector3df &position,
                                 core::vector3df &direction,
                                 core::vector3df &centrifugal_force,
                                 f32 t ) const
{ HH_ENTER
  // prepare
  const core::vector3df &P0 =  node.v_Position;
  const core::vector3df &P3 = _Data.v_Position;
  const core::vector3df  P1 = P0 +  node._Forward  *  node.ev_Forward;
  const core::vector3df  P2 = P3 + _Data._Backward * _Data.ev_Backward;
  // calculate some auxiliary points
  const core::vector3df A = t * (P1-P0);
  const core::vector3df B = t * (P2-P1);
  const core::vector3df C = t * (P3-P2);
  const core::vector3df D = t * ((P1+B)-(P0+A));
  const core::vector3df E = t * ((P2+C)-(P1+B));
  const core::vector3df T =     ((P1+B+E)-(P0+A+D));
  const core::vector3df P = t * T;
  // calculate the position
  position = P0 + A + D + P;
  // calculate the direction
  const f32 T_length = T.getLength();
  if ( T_length == 0 )
    direction.set( 0, 0, 0 );  // if ( P0 == P1 ) ...
  else
    direction = T / T_length;
  // calculate the vector of centrifugal force
  const core::vector3df K = T.crossProduct( T.crossProduct(position-P1) );
  const f32 K_length = K.getLength();
  if ( K_length == 0 )
  {
    centrifugal_force.set( 0, 0, 0 );  // if the actual point has no radius ...
    return false;
  }
  centrifugal_force = K / K_length;
  return true;
}


//! Fills up the given array by some positions of the way to this node position.
//! \param node The data of the previous node.
//! \param array The positions will be appended to this array.
//! \param range The approximated range of the points. Must be greater than zero!
void XNodePosition::getPositions( const XNodePositionData &node,
                                  vector3df_Array &array,
                                  f32 range ) const
{ HH_ENTER
  if ( range <= 0 )
    return;
  // calculate the number of points
  u32 no_of_points = (u32)(_Length / range);
  if ( no_of_points == 0 )
    no_of_points++;
  // iterate over all points (ignore the first point)
  core::vector3df position;
  for ( u32 i=1; i<=no_of_points; i++ )
  {
    // calculate the way
    const f32 way = _Length * (f32)i / (f32)no_of_points;
    // find the t
    const f32 t = getT( way );
    // get the position and append to the array
    getPosition( node, position, t );
    array.push_back( position );
  }
}


//! Fills up the given array by the node positions and their forward
//! and backward vectors.
//! 1. vector .. this is the node position.
//! 2. vector .. this is the forward vector multiplied by auxiliary range.
//! 3. vector .. this is the backward vector multiplied by auxiliary range.
//! \param array The positions and vectors will be appended to this array.
void XNodePosition::getPositionsAndDirections( vector3df_Array &array )
{ HH_ENTER
  array.push_back( _Data.v_Position );
  array.push_back( _Data._Forward  * _Data.ev_Forward );
  array.push_back( _Data._Backward * _Data.ev_Backward );
  HH_COUT( "_Data._Forward  = " << _Data._Forward );
  HH_COUT( "_Data._Backward = " << _Data._Backward );
}

                                               
//! Returns the time index at the given way length.
//! \param way The length of way.
f32 XNodePosition::getT( f32 way ) const
{ HH_ENTER
  if ( way < 0 )
    return 0;
  const f32 _dt = 1.0f / (f32)ANIMATOR_GRAVITION_BEZIER_CURVE_ELEMENTS;
  f32 _t = 0.0f;
  f32 _way = 0.0f;
  for ( int i=0; i<ANIMATOR_GRAVITION_BEZIER_CURVE_ELEMENTS; i++ )
  {
    const f32 old_way = _way;
    _way += _LengthOfElements[i];
    if ( _way >= way )
    {
      _t += ( _dt * (way-old_way) / _LengthOfElements[i] );
      break;
    }
    _t += _dt;
  }
  // end
  return ( (_t <= 1.0f) ? _t : 1.0f );
}


//! Initializes this node.
//! \param node The data of the previous node.
void XNodePosition::init( const XNodePositionData &node )
{ HH_ENTER
  if ( !_InitIsDone )
  {
    // prepare
    const core::vector3df &P0 =  node.v_Position;
    const core::vector3df &P3 = _Data.v_Position;
    const core::vector3df  P1 = P0 +  node._Forward  *  node.ev_Forward;
    const core::vector3df  P2 = P3 + _Data._Backward * _Data.ev_Backward;
    // divide the bezier curve into small peaces and calculate their length
    core::vector3df A, B=P0;
    f32 delta = 0;
    _Length = 0;
    for ( int i=0; i<ANIMATOR_GRAVITION_BEZIER_CURVE_ELEMENTS; i++ )
    {
      A = B;
      // calculate the next position
      if ( i == (ANIMATOR_GRAVITION_BEZIER_CURVE_ELEMENTS-1) )
        B = P3;
      else
      {
        const f32 _t   = (f32)(i+1) / (f32)ANIMATOR_GRAVITION_BEZIER_CURVE_ELEMENTS;
        const f32 _t2  = _t   * _t;
        const f32 _t3  = _t2  * _t;
        const f32 _1t  = 1.0f - _t;
        const f32 _1t2 = _1t  * _1t;
        const f32 _1t3 = _1t2 * _1t;
        B =                 _1t3 * P0 +
            (_t +_t +_t)  * _1t2 * P1 +
            (_t2+_t2+_t2) * _1t  * P2 +
                     _t3  *        P3;
      }
      // calculate the length
      _LengthOfElements[i] = (B-A).getLength();
      // add length to summary
      _Length += _LengthOfElements[i];
    }
    // ok
    _InitIsDone = true;
  }
}


//! Writes attributes of the scene node animator.
void XNodePosition::serializeAttributes
                    ( io::IAttributes *out,
                      io::SAttributeReadWriteOptions *options ) const
{ HH_ENTER
  out->addInt( "_Id", _Id );
  _Data.serializeAttributes( out, options );
}


/******************************************************************************/


//! Constructor.
XCameraSceneNodeAnimator::XCameraSceneNodeAnimator()
:ISceneNodeAnimator()
,_AutoReverse( false )
,_DefaultSpeed( 1.0f )
,_ForwardMoving( true )
,_IndexOfFirstNode( 0 )
,_IndexOfLastNode( 0xffffffffUL )
,_MasterSpeed( 1.0f )
,_Moving( true )
,_Repeat( false )
,v_Top( 0, 1, 0 )
,p_Camera( NULL )
,_PrivateNodeCounter( 0 )
{ HH_ENTER
#ifdef _DEBUG
  setDebugName( "XCameraSceneNodeAnimator" );
#endif
  // init.
  restart();
}


//! Destructor.
XCameraSceneNodeAnimator::~XCameraSceneNodeAnimator()
{ HH_ENTER
  // free all node positions
  clear();
  // free the camera
  setCamera();
}


//! Animates a scene node.
void XCameraSceneNodeAnimator::animateNode( ISceneNode *node, u32 timeMs )
{ HH_ENTER
  if ( !_Moving )
  { 
    _S._Time = timeMs;
    return;
  }
  // first init. all nodes if not already done
  init();
  // init. our private structure if any node has been appended
  if ( _PrivateNodeCounter != countNodes() )
  {
    _PrivateNodeCounter = countNodes();
    initS( node, timeMs );
  }
  // calculate the delta time
  u32 delta_time = timeMs - _S._Time;
  // calculate the new position and orientation
  while ( delta_time > 0 )
  { 
    const u32 used_time = calculateNextPositionAndOrientation( delta_time );
    delta_time -= used_time;
    _S._Time += used_time;
  }
  // set new position
  node->setPosition( _S._Position );
  // set new orientation
  if ( p_Camera )
    setOrientationOfCamera();
}


//! Changes the speed of the animator and all nodes by the given factor.
//! \param factor The speed will be multiplied by this factor. 
void XCameraSceneNodeAnimator::changeSpeed( f32 factor )
{ HH_ENTER
  // exit if negative
  if ( factor < 0.0 )
    return;
  // change defualt speed
  _DefaultSpeed *= factor;
  // change speed of all nodes
  for ( u32 i=0; i<countNodes(); i++ )
    if ( getNode(i)->_Data._Speed > 0 )
      getNode(i)->_Data._Speed *= factor;
}


//! Removes all node positions.
void XCameraSceneNodeAnimator::clear()
{ HH_ENTER
  // drop all node positions
  for ( u32 i=0; i<_Nodes.size(); i++ )
    _Nodes[i]->drop();
  // clear array
  _Nodes.clear();
  // restart animator
  restart();
}


//! Returns the number of nodes to be reached before stop moving.
//! If not moving a zero will be returned.
//! \return The number of nodes the animator has to reach vefore stop moving.
u32 XCameraSceneNodeAnimator::countNodesToBeReached() const
{ HH_ENTER
  if ( _Moving )
  {
    // if repeat is on we are moving endless
    if ( _Repeat )
      return 0xffffffffUL;
    // in which direction are we moving now?
    if ( _ForwardMoving )
    {
      // forward:
      return subIndex( getIndexOfLastNode(), getIndex() );
    }
    else
    {
      // backward:
      return subIndex( getIndex(), getIndexOfFirstNode() );
    }
  }
  // end
  return 0;
}


//! Creates a new node position. Use drop() to destroy this node.
//! \param position The position of this node.
//! \param forward_direction The forward direction.
//! The backward direction is set straight backwards. No edge!
XNodePosition *XCameraSceneNodeAnimator::createNodePosition( s32 id )
{ HH_ENTER
  XNodePosition *p_node = new XNodePosition(id);
  _Nodes.push_back( p_node );
  return p_node;
}


//! Reads attributes of the scene node animator.
void XCameraSceneNodeAnimator::deserializeAttributes
                               ( io::IAttributes *in,
                                 io::SAttributeReadWriteOptions *options )
{ HH_ENTER
  // clear old nodes
  clear();
  // read public data
  _AutoReverse      = in->getAttributeAsBool    (  "_AutoReverse" );
  _DefaultSpeed     = in->getAttributeAsFloat   (  "_DefaultSpeed" );
  _ForwardMoving    = in->getAttributeAsBool    (  "_ForwardMoving" );
  _IndexOfFirstNode = in->getAttributeAsInt     (  "_IndexOfFirstNode" );
  _IndexOfLastNode  = in->getAttributeAsInt     (  "_IndexOfLastNode" );
  _MasterSpeed      = in->getAttributeAsFloat   (  "_MasterSpeed" );
  _Moving           = in->getAttributeAsBool    (  "_Moving" );
  _Repeat           = in->getAttributeAsBool    (  "_Repeat" );
  v_Top             = in->getAttributeAsVector3d( "v_Top" );
  // read nodes
  XNodePosition::deserializeAttributes( in, options );  // this node ...
  for (;;)
  {
    const bool end_of_nodes = in->getAttributeAsBool( "EndOfNodes" );
    if ( end_of_nodes )
      break;
    XNodePosition *p_node = createNodePosition();
    p_node->deserializeAttributes( in, options );
  }
  // prepare to init.
  setCamera();  // may be no good idea to load a reference here ...
  // restart animator
  restart();
}


//! Returns the index of the node the animator is moving to.
//! Note that the result will be invalid if the animator is not moving.
//! \return The index of the next node. Zero for the animator.
u32 XCameraSceneNodeAnimator::getNextIndex() const
{ HH_ENTER
  u32 ret = getIndex();
  if ( _Moving )
  {
    // in which direction are we moving now?
    if ( _ForwardMoving )
    {
      // forward:
      const u32 next = addIndex( getIndex(), 1 );
      if ( !_Repeat )
      if ( ret == getIndexOfLastNode() )
        return ret;
      ret = next;
    }
    else
    {
      // backward:
      const u32 prev = subIndex( getIndex(), 1 );
      if ( !_Repeat )
      if ( ret == getIndexOfFirstNode() )
        return ret;
      ret = prev;
    }
  }
  // end
  return ret;
}


//! Returns the length of way.
//! \param index_from The index of the first point (zero==animator position)
//! \param index_to The index of the last point (zero==animator position)
//! \return The length from node 'index_from' to node 'index_to'.
//! Set 'index_to' equal to 'index_from' to get the length of cirlce.
//! Note that 'index_to' > 'index_from' returns the length of way in
//! case of a cirlce if '_Repeat' is set to on.
f32 XCameraSceneNodeAnimator::getLengthOfWay( u32 index_from, u32 index_to )
{ HH_ENTER
  f32 length = 0.0;
  // first init. all nodes if not already done
  init();
  // prepare
  index_from = index_from % countNodes();
  index_to   = index_to   % countNodes();
  // loop
  do
  {
    // next node
    index_from = (index_from+1) % countNodes();
    // add the length of way from the actual node to the next node
    const XNodePosition *p_node = getNode(index_from);
    length += p_node->_Length;
  } while ( index_from != index_to );
  // end
  return length;
}


//! Returns the address of a node using the given id.
//! \param id The id of the node.
//! \return The address of the node or NULL if not found.
XNodePosition *XCameraSceneNodeAnimator::getNodeById( s32 id )
{ HH_ENTER
  // loop over all nodes
  for ( u32 i=0; i<countNodes(); i++ )
  {
    XNodePosition *p_node = getNode(i);
    if ( id == p_node->_Id )
      return p_node; 
  }
  // end
  return NULL;
}


//! Fills up the given array by some positions of the way.
//! \param array The positions will be appended to this array.
//! \param range The approximated range of the points. Must be greater than zero!
//! \param index_from The index of the first point (zero==animator position)
//! \param index_to The index of the last point (zero==animator position)
void XCameraSceneNodeAnimator::getPositions( vector3df_Array &array,
                                             f32 range,
                                             u32 index_from,
                                             u32 index_to )
{ HH_ENTER
  array.clear();
  if ( range <= 0 )
    return;
  // init. if not already done
  init();
  if ( index_to >= countNodes() )
  {
    if ( _Repeat )
      index_to = countNodes();  // we need the way from last node to the animator too!
    else
      index_to = countNodes() - 1;  // stop at the last node!
  }
  // append first position if the ring is not closed
  const XNodePositionData *p_node = &refData( index_from );
  if ( (!_Repeat) ||
       ( (index_from == 0) &&
         ((index_to+1) == countNodes())
       )  
     )
  {
    array.push_back( p_node->v_Position );
  }
  // loop over all nodes
  for ( u32 i=index_from+1; i<=index_to; i++ )
  {
    getNode(i)->getPositions( *p_node, array, range );
    p_node = &refData(i);
  }
}


//! Fills up the given array by the node positions and their forward
//! and backward vectors.
//! 1. vector .. this is the node position.
//! 2. vector .. this is the forward vector multiplied by auxiliary range.
//! 3. vector .. this is the backward vector multiplied by auxiliary range.
//! Note that the first entry (zero index) stores all vectors of the
//! animator and the the backwar direction of the animator is undefined.
//! \param array The positions and vectors will be appended to this array.
//! \param index_from The index of the first point (zero==animator position)
//! \param index_to The index of the last point (zero==animator position)
void XCameraSceneNodeAnimator::getPositionsAndDirections( vector3df_Array &array,
                                                          u32 index_from,
                                                          u32 index_to )
{ HH_ENTER
  array.clear();
  // init. if not already done
  init();
  // loop over all nodes
  for ( u32 i=index_from; i<index_to && i<countNodes(); i++ )
    getNode(i)->getPositionsAndDirections( array );
}


//! Initializes all nodes. This may take some cpu time.
//! Call this method after changing any node before drawing your scene.
//! Note that animateNode() will call init() too.
void XCameraSceneNodeAnimator::init()
{ HH_ENTER
  XNodePositionData *p_node = &refData( countNodes()-1 );
  for ( u32 i=0; i<countNodes(); i++ )
  {
    getNode(i)->init( *p_node );
    p_node = &refData(i);
  }
}


//! Returns true if at least two following nodes are using zero speed.
//! In this case the animator will stop because zero speed means no moving.
//! \return true if the animator will stop at two zero speed nodes.
bool XCameraSceneNodeAnimator::isZeroSpeed() const
{ HH_ENTER
  bool is_zero = false;
  // loop over all nodes
  for ( u32 i=0; i<countNodes(); i++ )
    if ( getNode(i)->_Data.getSpeed(_DefaultSpeed) == 0 )
    {
      // check if this is the second zero speed
      if ( is_zero )
        return true;
      is_zero = true;
    }
    else
      is_zero = false;
  // end
  return false;
}


//! Writes attributes of the scene node animator.
void XCameraSceneNodeAnimator::serializeAttributes
                               ( io::IAttributes *out,
                                 io::SAttributeReadWriteOptions *options ) const
{ HH_ENTER
  // write public data
  out->addBool    (  "_AutoReverse",      _AutoReverse );
  out->addFloat   (  "_DefaultSpeed",     _DefaultSpeed );
  out->addBool    (  "_ForwardMoving",    _ForwardMoving );
  out->addInt     (  "_IndexOfFirstNode", _IndexOfFirstNode );
  out->addInt     (  "_IndexOfLastNode",  _IndexOfLastNode );
  out->addFloat   (  "_MasterSpeed",      _MasterSpeed );
  out->addBool    (  "_Moving",           _Moving );
  out->addBool    (  "_Repeat",           _Repeat );
  out->addVector3d( "v_Top",             v_Top );
  // write nodes
  XNodePosition::serializeAttributes( out, options );  // this node ...
  for ( u32 i=0; i<_Nodes.size(); i++ )
  {
    // use a boolean here to make it easy to edit a text file
    out->addBool( "EndOfNodes", false );
    _Nodes[i]->serializeAttributes( out, options );
  }
  out->addBool( "EndOfNodes", true );
}


//! Restarts the animator. The next position will be the position of start.
void XCameraSceneNodeAnimator::restart()
{ HH_ENTER
  init();
  _PrivateNodeCounter = 0;
}

// Checks if the animator has to stop moving.
// If a waypoint is reached the next waypoint will be initialized.
// Returns true while still moving.
bool XCameraSceneNodeAnimator::checkForStopMoving()
{ HH_ENTER
  // are we still moving?
  if ( !_Moving )
    return false;  // no!
  // is the actual node the last node of way?
  if ( getIndex() == getNextIndex() )  // never true if repeat is on (and possible)
  {
    // yes:
    // auto-reverse?
    if ( (_AutoReverse) &&
         (countNodes() > 1) )
    {
      // yes:
      _ForwardMoving = !_ForwardMoving;
      calculateAcceleration();
    }
    else
    {
      // no:
      _Moving = false;  // repeat is off
      return false;
    }
  }
  // end
  return true;
}


// Calculates the acceleration and moving time to the next node position.
// Call this method before starting to move to the next node.
void XCameraSceneNodeAnimator::calculateAcceleration()
{ HH_ENTER
  if ( _S._IndexOfNode < countNodes() )
  {
    // find the both node positions we are moving inside
    const XNodePositionData &dataA = refData( getIndex() );
    const XNodePositionData &dataB = refData( getNextIndex() );
    const f32 speedA = _S._MasterSpeed * dataA.getSpeed(_DefaultSpeed);
    const f32 speedB = _S._MasterSpeed * dataB.getSpeed(_DefaultSpeed);
    // get the way
    f32 way;
    if ( _ForwardMoving )
      way = nextNodePosition()->_Length;
    else
      way = getNode(getIndex())->_Length;
    // calculate the time
    const f32 speed = speedA + speedB;
    if ( speed <= 0 )
      _S._dTime = 0;
    else
      _S._dTime = (2.0 * way) / speed;  // average = 1/2 * (speedA+speedB)
    // calculate the acceleration
    if ( _S._dTime <= 0 )  // zero if no way
      _S._Acceleration = 0;
    else
      _S._Acceleration = (speedB - speedA) / _S._dTime;
  }
  else
  {
    _S._dTime = 0;
    _S._Acceleration = 0;
  }
  _S._Forward = _ForwardMoving;
  // we are starting at time index zero
  _S._aTime = 0;
  _S._T = 0;
}


// Calculates the next position.
u32 XCameraSceneNodeAnimator::calculateNextPositionAndOrientation( u32 delta_time )
{ HH_ENTER
  f32 dt = (f32)delta_time / 1000.0f;
  // check for end of moving
  if ( !checkForStopMoving() )
    return delta_time;  // all the time is used (we are waiting now)
  // correct direction if changed
  if ( _S._Forward != _ForwardMoving )
  {
    const bool begin_of_a_node = ( _S._aTime == 0 );
    // redirect some values
    _S._Acceleration = -_S._Acceleration;
    _S._aTime        = _S._dTime - _S._aTime;
    _S._Forward      = _ForwardMoving;
    // we need the next position of the old direction
    _ForwardMoving   = !_ForwardMoving;
    _S._IndexOfNode  = getNextIndex();
    _ForwardMoving   = _S._Forward;
    // at the begin of a node we must calculate new acceleration
    if ( begin_of_a_node )
      calculateAcceleration();
  }
  // find the both node positions we are moving inside
  const XNodePositionData &dataA = refData( getIndex() );
  const XNodePositionData &dataB = refData( getNextIndex() );
  if ( &dataA == &dataB )
    return delta_time;  // exit - we are not moving!
  const f32 speedA = _S._MasterSpeed * dataA.getSpeed(_DefaultSpeed);
  const f32 speedB = _S._MasterSpeed * dataB.getSpeed(_DefaultSpeed);
  // clip the time
  if ( (_S._aTime+dt) > _S._dTime )
    dt = _S._dTime - _S._aTime;
  if ( dt < 0.001f )
  {
    // next node
    _S._IndexOfNode = getNextIndex();
    _S._MasterSpeed = _MasterSpeed;  // get the actual master speed now
    calculateAcceleration();
    return 0;
  }
  delta_time = (u32)(dt*1000.0f+0.5f);
  _S._aTime += dt;
  // find the new position
  if ( _S._IndexOfNode < countNodes() )
  {
    // calculate the length of way
    f32 way = ( speedA * _S._aTime );
    if ( _S._Acceleration != 0 )
      way += ( 0.5f * _S._Acceleration * _S._aTime * _S._aTime );
    // calculate the new speed
    _S._Speed = speedA + _S._aTime * (speedB-speedA);
    // calculate the new position
    if ( p_Camera )
    {
      bool ok;
      if ( _ForwardMoving )
      {
        const XNodePosition *p_node = nextNodePosition();
        _S._T = p_node->getT( way );
        ok = p_node->getPosition( dataA, _S._Position,
                                         _S._Direction,
                                         _S._Centrifugal, _S._T );
      }
      else
      {
        const XNodePosition *p_node = getNode( getIndex() );
        _S._T = p_node->getT( p_node->_Length - way );
        ok = p_node->getPosition( dataB, _S._Position,
                                         _S._Direction,
                                         _S._Centrifugal, _S._T );
      }
      if ( !ok )
      {
        _S._Centrifugal = -v_Top;
        _S._Centrifugal.normalize();
      }
    }
    else
    {
      if ( _ForwardMoving )
      {
        const XNodePosition *p_node = nextNodePosition();
        _S._T = p_node->getT( way );
        p_node->getPosition( dataA, _S._Position, _S._T );
      }
      else
      {
        const XNodePosition *p_node = getNode( getIndex() );
        _S._T = p_node->getT( p_node->_Length - way );
        p_node->getPosition( dataB, _S._Position, _S._T );
      }
      _S._Centrifugal = -v_Top;
      _S._Centrifugal.normalize();
    }
    _S._Top = v_Top;
  }
  // end
  return delta_time;
}


// Initializes _S.
void XCameraSceneNodeAnimator::initS( ISceneNode *node, u32 timeMs )
{ HH_ENTER
  // acceleration (constant betweed nodes)
  _S._Acceleration = 0.0;
  // the delta time of _T
  _S._aTime = 0.0;
  // vector of centrifugal force
  _S._Centrifugal = -v_Top;
  // vector of direction (may be zero)
  _S._Direction.set( 0, 0, 0 );
  // the time of moving to the next node
  _S._dTime = 0.0;
  // true=forward, false=backward
  _S._Forward = _ForwardMoving;
  // the index of the last reached node
  if ( _AutoReverse )
    _S._IndexOfNode = getIndexOfLastNode();
  else
    _S._IndexOfNode = getIndexOfFirstNode();
  // the master speed of last reached node
  _S._MasterSpeed = _MasterSpeed;
  // aktual Position
  _S._Position = node->getPosition();
  // true while running (false==restart)
  _S._Running = true;
  // the actual speed
  _S._Speed = _S._MasterSpeed * _DefaultSpeed;
  // index [0,1] between two nodes
  _S._T = 0.0;
  // the time of last frame
  _S._Time = timeMs;
  // vector showing the top of orientation
  _S._Top = v_Top;
  // now we need some initial values of our first node
  XNodePositionData &data = refData( _S._IndexOfNode );
  _S._Position = data.v_Position;
  _S._Speed = _S._MasterSpeed * data.getSpeed( _DefaultSpeed );
  // at last we need the acceleration and time to move to the next node
  calculateAcceleration();
}


// Set the orientation of the camera.
void XCameraSceneNodeAnimator::setOrientationOfCamera()
{ HH_ENTER
  // prepare
  const XNodePositionData &dataA = refData( getIndex() );
  const XNodePositionData &dataB = refData( getNextIndex() );
  core::vector3df posB, topB;
  // same orientation?
  if ( dataA._Orientation == dataB._Orientation )
  {
    bool ok = false;
    // yes:
    switch ( dataB._Orientation )
    {
      case XNodePositionData::OT_FIXPOINT:
      { if ( dataA.v_Fixpoint == dataB.v_Fixpoint )
        {
          posB = dataB.v_Fixpoint;
          topB = v_Top;
          ok = true;
        }
      } break;
      case XNodePositionData::OT_FORWARD:
      { posB = _S._Position + _S._Direction;
        topB = _S._Top;
        ok = true;
      } break;
      case XNodePositionData::OT_VIEWPOINT:
      { if ( dataA.getViewpoint() == dataB.getViewpoint() )
        {
          posB = dataB.getViewpoint() + dataB.v_Fixpoint;
          topB = v_Top;
          ok = true;
        }
      } break;
    }
    // set camera and exit
    if ( ok )
    {
      p_Camera->setTarget( posB );
      p_Camera->setUpVector( topB );
      return;
    }
  }
  // we need to do interpolation
  core::vector3df posA, topA;
  core::vector3df dir, cent;
  // get the orientation of the last node
  if ( _ForwardMoving )
  {
    const XNodePosition *p_node = nextNodePosition();
    p_node->getPosition( dataA, posA, dir, cent, 0 );
  }
  else
  {
    const XNodePosition *p_node = getNode( getIndex() );
    p_node->getPosition( dataB, posA, dir, cent, 1 );
  }
  topA = v_Top;
  switch ( dataA._Orientation )
  {
    case XNodePositionData::OT_FIXPOINT:
    { posA = dataA.v_Fixpoint - posA;
    } break;
    case XNodePositionData::OT_FORWARD:
    { posA = dir; 
    } break;
    case XNodePositionData::OT_VIEWPOINT:
    { posA = dataA.getViewpoint() + dataA.v_Fixpoint - posA;
    } break;
  }
  posA.normalize();
  // get the orientation of the next node
  if ( _ForwardMoving )
  {
    const XNodePosition *p_node = nextNodePosition();
    p_node->getPosition( dataA, posB, dir, cent, 1 );
  }
  else
  {
    const XNodePosition *p_node = getNode( getIndex() );
    p_node->getPosition( dataB, posB, dir, cent, 0 );
  }
  topB = v_Top;
  switch ( dataB._Orientation )
  {
    case XNodePositionData::OT_FIXPOINT:
    { posB = dataB.v_Fixpoint - posB;
    } break;
    case XNodePositionData::OT_FORWARD:
    { posB = dir;
    } break;
    case XNodePositionData::OT_VIEWPOINT:
    { posB = dataB.getViewpoint() + dataB.v_Fixpoint - posB;
    } break;
  }
  posB.normalize();
  // calculate and set the actual oriention
  const f32 interpolate = ( (_S._dTime!=0) ? (_S._aTime/_S._dTime) : 0.0 );
  const core::vector3df posD = ( posB - posA );
  posB = _S._Position + posA + interpolate * posD;
  p_Camera->setTarget( posB );
  p_Camera->setUpVector( topB );
}


} // end namespace scene
} // end namespace irr

