// Copyright (C) 2008 Helmut Hirtes
// This file is a free extension of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#ifndef __X_ANIMATORCAMERASCENENODE_H_INCLUDED__
#define __X_ANIMATORCAMERASCENENODE_H_INCLUDED__


#include "ICameraSceneNode.h"
#include "ISceneNode.h"
#include "irrArray.h"


namespace irr
{
namespace scene
{


using namespace irr;


  // Prototypes
  class XCameraSceneNodeAnimator;


  // The number of peaces of a full bezier curve
  #define ANIMATOR_GRAVITION_BEZIER_CURVE_ELEMENTS    (32)


  //! The data of a node.
  class XNodePositionData
  {

    public:

      //! The type of orientation.
      enum eT_Orientation
      { OT_FIXPOINT,    // looking to a fixpoint
        OT_FORWARD,     // the top is given by a vector
        OT_VIEWPOINT    // looking to another scene node using getPosition()+offset
      };

      //! The range of the incoming auxiliary point.
      //! Must be greater than 0.0
      f32 _Backward;

      //! Vektor of length 1.0 in backward direction.
      //! This is the incoming direction.
      core::vector3df ev_Backward;

      //! The fixpoint (OT_FIXPOINT) also used as offset (OT_VIEWPOINT).
      //! Default=( 0, 0, 0 )
      core::vector3df v_Fixpoint;

      //! The range of the outgoing auxiliary point.
      //! Must be greater than 0.0
      f32 _Forward;

      //! Vektor of length 1.0 in forward direction.
      //! This is the outgoing direction.
      core::vector3df ev_Forward;

      //! A pointer to another scene node used as viewpoint or NULL.
      //! See OT_VIEWPOINT.
      ISceneNode *p_Viewpoint;

      //! The type of orientation. Default=OT_FORWARD
      eT_Orientation _Orientation;

      //! The position of this node.
      core::vector3df v_Position;

      //! The speed at this position.
      //! If negative the speed is given by the animator as default speed.
      f32 _Speed;

    public:

      //! Constructor.
      XNodePositionData()
      :_Backward( 1 )
      ,v_Fixpoint( 0, 0, 0 )
      ,_Forward( 1 )
      ,p_Viewpoint( NULL )
      ,_Orientation( OT_FORWARD )
        { clearSpeed(); }

      //! Destructor.
      ~XNodePositionData()
        { setViewpoint( NULL ); }

      //! Clear the speed. Now the default speed given by animator is used.
      void clearSpeed()
        { _Speed = -1; }

      //! Reads attributes of the scene node animator.
      virtual void deserializeAttributes( io::IAttributes *in, io::SAttributeReadWriteOptions *options=0 );

      //! Returns the speed.
      f32 getSpeed( f32 default_speed ) const
        { return ( (_Speed < 0) ? default_speed : _Speed ); }

      //! Returns the viewpoint or (0,0,0) if not existing.
      core::vector3df getViewpoint() const
        { if ( p_Viewpoint != NULL )
            return p_Viewpoint->getPosition();
          return core::vector3df( 0, 0, 0 );
        }

      //! Writes attributes of the scene node animator.
      virtual void serializeAttributes( io::IAttributes *out, io::SAttributeReadWriteOptions *options=0 ) const;

      //! Set backward vektor (incoming direction).
      void setBackward( const core::vector3df &direction )
        { ev_Backward = direction;
          ev_Backward.normalize();
        }

      //! Set forward vektor (outgoing direction).
      void setForward( const core::vector3df &direction )
        { ev_Forward = direction;
          ev_Forward.normalize();
        }

      //! Set position.
      void setPosition( const core::vector3df &position )
        { v_Position = position; }

      //! Set the speed at the position of this node.
      void setSpeed( f32 speed )
        { if ( speed >= 0.0 )
            _Speed = speed;
        }

      //! Set a scene node as point of view.
      //! \param p_scenenode The address of the scene node or NULL.
      void setViewpoint( ISceneNode *p_scenenode=NULL )
        { if ( p_Viewpoint != NULL )
            p_Viewpoint->drop();
          p_Viewpoint = p_scenenode;
          if ( p_Viewpoint != NULL )
            p_Viewpoint->grab();
        }

  };


  // A type of array of class core::vector3df:
  typedef core::irrAllocatorFast<core::vector3df> vector3df_Allocator;
  typedef core::array<core::vector3df,vector3df_Allocator> vector3df_Array;


  //! A position of the scene node gravitation animator.
  //! Note that a node defines a state to be reached at the position of this node.
  //! While moving from previous node to this node this state will be reached.
  //! The time doing this depends on the time of moving from previous node to this node.
  class XNodePosition : virtual public IReferenceCounted
  {

    public:

      //! This friend class is allowed to access protected methods and data.
      friend class XCameraSceneNodeAnimator;

      //! This id is used to identify this node.
      s32 _Id;

      //! Constructor.
      //! \param id This id is used to identify this node.
      XNodePosition( s32 id=-1 );


      //! Clear the speed. Now the default speed given by animator is used.
      void clearSpeed()
        { _Data.clearSpeed(); }

      //! Reads attributes of the scene node animator.
      virtual void deserializeAttributes( io::IAttributes *in, io::SAttributeReadWriteOptions *options=0 );

      //! Get the range of the auxiliary points.
      //! \param backward The range of the backward auxiliary point.
      //! \param forward The range of the forward auxiliary point.
      void getAuxiliary( f32 backward, f32 forward ) const
        { backward = _Data._Backward;
          forward  = _Data._Forward;
        }

      //! Returns the backward (incoming) direction.
      const core::vector3df &getBackwardDirection() const
        { return _Data.ev_Backward; }

      //! Returns the forward (outgoing) direction.
      const core::vector3df &getForwardDirection() const
        { return _Data.ev_Forward; }

      //! Returns the position of this node.
      const core::vector3df &getPosition() const
        { return _Data.v_Position; }

      //! Writes attributes of the scene node animator.
      virtual void serializeAttributes( io::IAttributes *out, io::SAttributeReadWriteOptions *options=0 ) const;

      //! Set the range of the auxiliary points.
      //! These ranges may be used like the radius of a circle or ellipse.
      //! If smaller the curve looks like a edge. If greater you will get a loop. 
      //! \param backward The range of the backward auxiliary point. Must be greater than zero!
      //! \param forward The range of the forward auxiliary point. Must be greater than zero!
      void setAuxiliary( f32 backward, f32 forward )
        { _Data._Backward = ( (backward <= 0) ? 1 : backward );
          _Data._Forward  = ( (forward  <= 0) ? 1 : forward );
        }

      //! Set the backward (incoming) direction.
      //! \param direction The direction (incoming) of the tangent.
      void setBackwardDirection( const core::vector3df &direction )
        { _Data.setBackward( direction );
          _InitIsDone = false;
        }

      //! Set the backward (incoming) and forward (outgoing) direction.
      //! Do not set direction equal to the camera top because in this case
      //! the camera has no direction and you may see a fast camera turn.
      //! \param direction The direction (outgoing) of the tangent.
      void setDirection( const core::vector3df &direction )
        { _Data.setForward( direction );
          _Data.setBackward( -direction );
          _InitIsDone = false;
        }

      //! Set the backward (outgoing) direction.
      //! \param direction The direction (outgoing) of the tangent.
      void setForwardDirection( const core::vector3df &direction )
        { _Data.setForward( direction );
          _InitIsDone = false;
        }

      //! Set the orientation of the (optional) given camera.
      //! Here the orientation points to the given fixpoint.
      //! \param position The fixed position.
      void setOrientationToFixpoint( core::vector3df position )
        { _Data.v_Fixpoint = position;
          _Data._Orientation = XNodePositionData::OT_FIXPOINT;
        }

      //! Set the orientation of the (optional) given camera.
      //! Here the orientation points into forward moving direction.
      void setOrientationToForward()
        { _Data._Orientation = XNodePositionData::OT_FORWARD; }

      //! Set the orientation of the (optional) given camera.
      //! Here the orientation points to the position of the given
      //! scene node. Optional a offset can be given.
      //! \param p_scenenode The scene node.
      //! \param offset The offset.
      void setOrientationToViewpoint( ISceneNode *p_scenenode,
                                      core::vector3df offset=(0,0,0) )
        { _Data.setViewpoint( p_scenenode );
          _Data.v_Fixpoint = offset;
          _Data._Orientation = XNodePositionData::OT_VIEWPOINT;
        }

      //! Set the position of this node.
      //! \param position The position of this node.
      void setPosition( const core::vector3df &position )
        { _Data.setPosition( position );
          _InitIsDone = false;
        }

      //! Set the speed at the position of this node.
      void setSpeed( f32 speed )
        { _Data.setSpeed( speed ); }

    protected:

      //! Calculates the position between the last node and this node.
      //! \param node The data of the previous node.
      //! \param position Is set to the calcualted position.
      //! \param t The relative position (range from 0.0 to 1.0).
      void getPosition( const XNodePositionData &node,
                        core::vector3df &position,
                        f32 t ) const;

      //! Calculates the position between the last node and this node.
      //! Note that the vector of direction will be zero if both node
      //! position are equal.
      //! Note that the vector of centrifugal force may be a zero vector in
      //! the case that the actual position has no radius.
      //! \param node The data of the previous node.
      //! \param position Is set to the calcualted position.
      //! \param direction The vector of direction.
      //! \param centrifugal_force The vector of centrifugal force.
      //! \param t The relative position (range from 0.0 to 1.0).
      //! \return true if the vector of centrifugal force is not a zero vector.
      bool getPosition( const XNodePositionData &node,
                        core::vector3df &position,
                        core::vector3df &direction,
                        core::vector3df &centrifugal_force,
                        f32 t ) const;

      //! Fills up the given array by some positions of the way to this node position.
      //! \param node The data of the previous node.
      //! \param array The positions will be appended to this array.
      //! \param range The approximated range of the points. Must be greater than zero!
      void getPositions( const XNodePositionData &node,
                         vector3df_Array &array,
                         f32 range ) const;

      //! Fills up the given array by the node positions and their forward
      //! and backward vectors.
      //! 1. vector .. this is the node position.
      //! 2. vector .. this is the forward vector multiplied by auxiliary range.
      //! 3. vector .. this is the backward vector multiplied by auxiliary range.
      //! \param array The positions and vectors will be appended to this array.
      void getPositionsAndDirections( vector3df_Array &array );

      //! Returns the time index at the given way length.
      //! \param way The length of way.
      f32 getT( f32 way ) const;

      //! Initialize this node.
      //! \param node The data of the previous node.
      void init( const XNodePositionData &node );

    protected:

      //! The Data of this node.
      XNodePositionData _Data;

      //! Is set to true if this node is already initialized.
      bool _InitIsDone;

      //! The length of the way. This value is set by init().
      f32 _Length;

    private:

      // The length of peaces of the bezier curve.
      f32 _LengthOfElements[ANIMATOR_GRAVITION_BEZIER_CURVE_ELEMENTS];

  };


  // A type of array of class XNodePosition:
  typedef core::irrAllocatorFast<XNodePosition*> XNodePosition_Allocator;
  typedef core::array<XNodePosition*,XNodePosition_Allocator> XNodePosition_Array;


	//! Scene node gravitation animator.
  //! If you animate a unvisible scene node you can attach your camera
  //! to this scene node. This effect is like a look to something.
  //! You can change the public control variables while animator is moving.
  //! Note that changing values (e.g. speed) will immediately effect the animator. 
	class XCameraSceneNodeAnimator : virtual public ISceneNodeAnimator,
                                   public XNodePosition
	{

    public:

      //! Set to true to change moving direction at the end of a way.
      //! Note that '_Repeat' must be off.
      //! Default=false
      bool _AutoReverse;

      //! The default speed of all nodes if no speed is specified.
      //! This speed will be affected by _MasterSpeed.
      //! Do not set a negative speed.
      //! Default=1.0
      f32 _DefaultSpeed;

      //! Set to true to move forward else set to false to move backward.
      //! Default=true
      bool _ForwardMoving;

      //! Is set to the index of the first node.
      //! Set to zero to start at the position of the animator.
      //! Note that this index shows the last node while moving backwards.
      //! Default=0
      u32 _IndexOfFirstNode;

      //! Is set to the index of the last node.
      //! Set to max. to never stop.
      //! Note that this index shows the first node while moving backwards.
      //! Default=0xffffffffUL
      u32 _IndexOfLastNode;

      //! A master speed factor to change speed without changing the speed
      //! of the nodes. Note that the '_DefaultSpeed' will be affected too.
      //! Note that the master speed will be used after the next node was
      //! reached only. This avoids jump movements.
      //! Do not set a negative speed.
      //! Default=1.0
      f32 _MasterSpeed;

      //! Set this to false to stop moving.
      //! Default=true
      bool _Moving;

      //! Set to true to repeat moving endless.
      //! Default=false
      bool _Repeat;

      //! The direction of the camera top.
      //! Default=( 0, 1, 0 )
      core::vector3df v_Top;

    public:

      //! Constructor.
      XCameraSceneNodeAnimator(); 

      //! Destructor.
      virtual ~XCameraSceneNodeAnimator(); 

      //! Animates a scene node.
      virtual void animateNode( ISceneNode *node, u32 timeMs );

      //! Changes the speed of the animator and all nodes by the given factor.
      //! \param factor The speed will be multiplied by this factor. 
      void changeSpeed( f32 factor );

      //! Removes all node positions.
      void clear();

      //! Returns the number of stored nodes.
      //! \return The number of nodes. Note that the animator is the first node.
      u32 countNodes() const
        { return ( 1 + _Nodes.size() ); }

      //! Returns the number of nodes to be reached before stop moving.
      //! If not moving a zero will be returned.
      //! \return The number of nodes the animator has to reach vefore stop moving.
      u32 countNodesToBeReached() const;

      //! Creates a new node position. Use drop() to destroy this node.
      //! \return The address of the node.
      XNodePosition *createNodePosition( s32 id=-1 );

      //! Reads attributes of the scene node animator.
      virtual void deserializeAttributes( io::IAttributes *in, io::SAttributeReadWriteOptions *options=0 );

      //! Returns the index of the last reached node.
      //! \return The index of the last reached node. Zero for the animator.
      u32 getIndex() const
        { return _S._IndexOfNode; }

      //! Returns the index of the node the animator is moving to.
      //! Note that the result will be invalid if the animator is not moving.
      //! \return The index of the next node. Zero for the animator.
      u32  getNextIndex() const;

      //! Returns the length of way.
      //! \param index_from The index of the first point (zero==animator position)
      //! \param index_to The index of the last point (zero==animator position)
      //! \return The length from node 'index_from' to node 'index_to'.
      //! Set 'index_to' equal to 'index_from' to get the length of cirlce.
      //! Note that 'index_to' > 'index_from' returns the length of way in
      //! case of a cirlce if '_Repeat' is set to on.
      f32 getLengthOfWay( u32 index_from=0, u32 index_to=0 );

      //! Returns the address of a node. Note that the animator is the first node.
      //! \param index The index of the node. Range from 0 to countNodes()-1.
      //! Note that the modulo operator is used here. Index overflow is allowed.
      //! \return The address of the node.
      XNodePosition *getNode( u32 index )
        { index = index % countNodes();
          if ( index > 0 )
            return _Nodes[index-1];
          return this;
        }

      //! Returns the address of a node using the given id. 
      //! \param id The id of the node.
      //! \return The address of the node or NULL if not found.
      XNodePosition *getNodeById( s32 id );

      //! Returns the address of a node. Note that the animator is the first node.
      //! \param index The index of the node. Range from 0 to countNodes()-1.
      //! Note that the modulo operator is used here. Index overflow is allowed. 
      //! \return The address of the node.
      const XNodePosition *getNode( u32 index ) const
        { index = index % countNodes();
          if ( index > 0 )
            return _Nodes[index-1];
          return this;
        }

      //! Fills up the given array by some positions of the way.
      //! \param array The positions will be appended to this array.
      //! \param range The approximated range of the points. Must be greater than zero!
      //! \param index_from The index of the first point (zero==animator position)
      //! \param index_to The index of the last point (zero==animator position)
      void getPositions( vector3df_Array &array,
                         f32 range=1e+38f,
                         u32 index_from=0,
                         u32 index_to=0xffffffffUL );

      //! Fills up the given array by the node positions and their forward
      //! and backward vectors.
      //! 1. vector .. this is the node position.
      //! 2. vector .. this is the forward vector multiplied by auxiliary range.
      //! 3. vector .. this is the backward vector multiplied by auxiliary range.
      //! Note that the first entry (zero index) stores all vectors of the
      //! animator and the the backwar direction of the animator is undefined.
      //! \param array The positions and vectors will be appended to this array.
      //! \param index_from The index of the first point (zero==animator position)
      //! \param index_to The index of the last point (zero==animator position)
      void getPositionsAndDirections( vector3df_Array &array,
                                      u32 index_from=0,
                                      u32 index_to=0xffffffffUL );

      //! Initializes all nodes. This may take some cpu time.
      //! Call this method after changing any node before drawing your scene.
      //! Note that animateNode() will call init() too.
      void init();

      //! Returns true if at least two following nodes are using zero speed.
      //! In this case the animator will stop because zero speed means no moving.
      //! \return true if the animator will stop at two zero speed nodes.
      bool isZeroSpeed() const;

      //! Writes attributes of the scene node animator.
      virtual void serializeAttributes( io::IAttributes *out, io::SAttributeReadWriteOptions *options=0 ) const;

      //! Set or clear the address of a camera.
      //! This camery will look into the direction of moving.
      //! Note that this camera should not be controllable by user.
      //! \param camera The address of the camera or NULL.
      void setCamera( ICameraSceneNode *camera=NULL )
        { if ( p_Camera )
            p_Camera->drop();
          p_Camera = camera;
          if ( p_Camera )
            p_Camera->grab();
        }

      //! Restarts the animator. The next position will be the position of start.
      void restart();

      //! Set the default speed used by all nodes that do not specify their own speed.
      void setDefaultSpeed( f32 speed )
        { if ( speed >= 0 )
            _DefaultSpeed = speed;
          else
            _DefaultSpeed = 0;
        }

      //! Returns type of the scene node animator
      //! \return The type of animator.
      virtual ESCENE_NODE_ANIMATOR_TYPE getType() const
        { return ESNAT_FOLLOW_SPLINE; }  // ### tbd - may be changed later!

    private:

      // The address of a camera or NULL.
      // This camery will look into the direction of moving.
      ICameraSceneNode *p_Camera;

      // Array of node positions.
      XNodePosition_Array _Nodes;

      // A counter of nodes (to find time to init)
      u32 _PrivateNodeCounter;

      // A structure of the actual animation data.
      struct
      { public:
        f32             _Acceleration; // acceleration (constant between nodes)
        f32             _aTime;        // the time of position depending on way length
        core::vector3df _Centrifugal;  // vector of centrifugal force
        core::vector3df _Direction;    // vector of direction (may be zero)
        f32             _dTime;        // the time of moving to the next node
        bool            _Forward;      // true=forward, false=backward
        u32             _IndexOfNode;  // the index of the last reached node
        f32             _MasterSpeed;  // the master speed of last reached node
        core::vector3df _Position;     // aktual Position
        bool            _Running;      // true while running (false==restart)
        f32             _Speed;        // the actual speed
        f32             _T;            // index [0,1] between two nodes
        u32             _Time;         // the time of last frame
        core::vector3df _Top;          // vector showing the top of orientation
      } _S;

      // Adds a positive range to the given index.
      // The result will be inside the allowed range.
      u32 addIndex( u32 index, u32 add ) const
        { return ( (index+add) % countNodes() ); }

      // Checks if the animator has to stop moving.
      // If a waypoint is reached the next waypoint will be initialized.
      // Returns true while still moving.
      bool checkForStopMoving();

      // Calculates the acceleration and moving time to the next node position.
      // Call this method before starting to move to the next node.
      void calculateAcceleration();

      // Calculates the next position.
      // Returns the left time. This time may be greate zero in that case
      // we reached a node position. Must never return zero!
      u32 calculateNextPositionAndOrientation( u32 delta_time );

      // Returns the first/last node.
      u32 getIndexOfFirstNode() const
        { return ( _IndexOfFirstNode % countNodes() ); }
      u32 getIndexOfLastNode() const
        { return ( (_IndexOfLastNode < countNodes()) ? _IndexOfLastNode :
                                                       countNodes()-1 ); }

      // Initializes _S.
      void initS( ISceneNode *node, u32 timeMs );

      // Returns the next node.
      const XNodePosition *nextNodePosition()
        { return getNode( getNextIndex() ); }

      // Returns the reference of a node position data structure.
      // If the index is out of range the data structure of the animator is returned.
      XNodePositionData &refData( u32 index )
        { return getNode(index)->_Data; }

      // Set the orientation of the camera.
      void setOrientationOfCamera();

      // Subtracts a positive range from the given index.
      // The result will be inside the allowed range.
      u32 subIndex( u32 index, u32 sub ) const
        { return ( (countNodes()+index-sub) % countNodes() ); }

  };

  
} // end namespace scene
} // end namespace irr

#endif

