// Copyright (C) 2008 Helmut Hirtes
// This file is a free extension of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#include "XEventReceiver.h"


namespace irr
{
namespace nth
{


//! The identification of new user events.
//! Default=-1
s32 XEventReceiver::_UserData1 = -1;


//! Clears the list of event receiver.
void XEventReceiver::Clear()
{
  for ( u32 i=0; i<_EventReceiver.size(); i++ )
    _EventReceiver[i]->drop();
  _EventReceiver.clear();
}


//! Searches for a event receiver.
//! \param id The identitification number of the event receiver.
//! \return The address of the event receiver or NULL if not found.
XEventReceiver *XEventReceiver::Find( s32 id ) const
{
  for ( u32 i=0; i<_EventReceiver.size(); i++ )
  if ( id == _EventReceiver[i]->_Id )
    return _EventReceiver[i];
  return NULL;
}


//! Insert a new event receiver.
//! Do not change '_Priority' after inserting.
void XEventReceiver::Insert( XEventReceiver *p_xer )
{
  if ( p_xer )
  if ( p_xer != this )
  {
    p_xer->grab();
    for ( u32 i=0; i<_EventReceiver.size(); i++ )
    if ( p_xer->_Priority > _EventReceiver[i]->_Priority )
    {
      _EventReceiver.insert( p_xer, i );
      return;
    }
    _EventReceiver.push_back( p_xer );
  }
}


//! Called if an event happened.
//! \return true if the event was processed.
bool XEventReceiver::OnEvent( const SEvent &event )
{
  if ( _Active )
  {
    switch ( event.EventType )
    {
      case EET_GUI_EVENT:
      { if ( OnGuiEvent(event.GUIEvent) )
          return true;
      } break;
      case EET_KEY_INPUT_EVENT:
      { if ( OnKeyEvent(event.KeyInput) )
          return true;
      } break;
      case EET_LOG_TEXT_EVENT:
      { if ( OnLogEvent(event.LogEvent) )
          return true;
      } break;
      case EET_MOUSE_INPUT_EVENT:
      { if ( OnMouseEvent(event.MouseInput) )
          return true;
      } break;
      case EET_USER_EVENT:
      { if ( event.UserEvent.UserData1 == _UserData1 )
        { if ( OnUserCodeEvent(event.UserEvent.UserData2) )
            return true;
        }
        else
        { if ( OnUserEvent(event.UserEvent) )
            return true;
        }
      } break;
    }
    // give event to the event receiver of the list
    for ( u32 i=0; i<_EventReceiver.size(); i++ )
      if ( _EventReceiver[i]->OnEvent(event) )
        return true;
  }
  // end
  return false;
}


//! Remove the given event receiver.
void XEventReceiver::Remove( XEventReceiver *p_xer )
{
  if ( p_xer )
  for ( u32 i=0; i<_EventReceiver.size(); i++ )
  if ( p_xer == _EventReceiver[i] )
  {
    _EventReceiver.erase( i );
    p_xer->drop();
    break;
  }
}


//! Send a new user defined event.
//! \param code The event parameter 'UserData2'.
//! 'UserData1' is set to '_UserData1' to differ from old user events.
//! \return true if the event was handled.
bool XEventReceiver::SendUserEvent( s32 code )
{
  SEvent event;
  event.EventType = EET_USER_EVENT;
  event.UserEvent.UserData1 = _UserData1;
  event.UserEvent.UserData2 = code;
  event.UserEvent.UserData3 = core::PI;
  return OnEvent( event );
}


//! Send a new user defined event.
//! \param codes A list of event parameters 'UserData2'.
//! See SendUserEvent(s32) too.
//! \param from The index of the first user event code.
//! \param to The index of the last user event code.
void XEventReceiver::SendUserEvent( s32_Array codes, u32 from, u32 to )
{
  for ( u32 i=from; i<codes.size() && i<=to; i++ )
    SendUserEvent( codes[i] );
}


//! (De-)activates a list of keys.
//! \param p_keylist A list of keys. Letters are case insensitive.
//! Note that characters are ignored if there does no key code exist.
//! \param activate Set to false to deactivate the given keys.
void XExampleEventReceiver::installKeys( const char *p_keylist, bool activate )
{
  if ( p_keylist )
  {
    for (;;)
    {
      char chr = *(p_keylist++);
      if ( chr == 0x00 )
        break;
      // convert to upper case
      if ( (chr >= 'a') &&
           (chr <= 'z') )
        chr = (char)(chr - 'a' + 'A');
      // letters
      if ( (chr >= 'A') &&
           (chr <= 'Z') )
        HandleKey[KEY_KEY_A+chr-'A'] = activate;
      else
      // digits
      if ( (chr >= '0') &&
           (chr <= '9') )
        HandleKey[KEY_KEY_0+chr-'0'] = activate;
      else
      // others
      switch ( chr )
      {
        case 0x08:
        { HandleKey[KEY_BACK] = activate;
        } break;
        case 0x09:
        { HandleKey[KEY_TAB] = activate;
        } break;
        case 0x0C:
        { HandleKey[KEY_CLEAR] = activate;
        } break;
        case 0x0D:
        { HandleKey[KEY_RETURN] = activate;
        } break;
        case 0x1B:
        { HandleKey[KEY_ESCAPE] = activate;
        } break;
        case ' ':
        { HandleKey[KEY_SPACE] = activate;
        } break;
        case '*':
        { HandleKey[KEY_MULTIPLY] = activate;
        } break;
        case '+':
        { HandleKey[KEY_ADD] = activate;
        } break;
        case '-':
        { HandleKey[KEY_SUBTRACT] = activate;
        } break;
        case '/':
        { HandleKey[KEY_DIVIDE] = activate;
        } break;
      }
    }
  }
}


//! (De-)activates a list of special keys.
//! \param p_keylist A list of special keys. Special keys are e.g.
//! digits mapped to the num pad and letters mapped to the F1-F24 keys. 
//! '?' is used for the help key. 
//! Note that characters are ignored if there does no key code exist.
//! \param activate Set to false to deactivate the given keys.
void XExampleEventReceiver::installSpecialKeys( const char *p_keylist, bool activate )
{
  if ( p_keylist )
  {
    for (;;)
    {
      char chr = *(p_keylist++);
      if ( chr == 0x00 )
        break;
      // convert to upper case
      if ( (chr >= 'a') &&
           (chr <= 'z') )
        chr = (char)(chr - 'a' + 'A');
      // F1-F24 keys
      if ( (chr >= 'A') &&
           (chr <= 'Y') )
        HandleKey[KEY_F1+chr-'A'] = activate;
      else
      // digits (num pad)
      if ( (chr >= '0') &&
           (chr <= '9') )
        HandleKey[KEY_NUMPAD0+chr-'0'] = activate;
      else
      // others
      switch ( chr )
      {
        case '?':
        { HandleKey[KEY_HELP] = activate;
        } break;
      }
    }
  }
}


} // end namespace nth
} // end namespace irr
