// Copyright (C) 2008 Helmut Hirtes
// This file is a free extension of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#ifndef __X_EVENT_RECEIVER_H_INCLUDED__
#define __X_EVENT_RECEIVER_H_INCLUDED__


#include "XNiceToHave.h"


namespace irr
{
namespace nth
{


  //! Class of a extended event receiver.
  class XEventReceiver : virtual public IReferenceCounted,
                                 public IEventReceiver
  {

    public:

      // A type of array of user defined event codes.
      typedef core::irrAllocatorFast<s32> s32_Allocator;
      typedef core::array<s32,s32_Allocator> s32_Array;

      // A type of array of class XEventReceiver.
      typedef core::irrAllocatorFast<XEventReceiver*> XEventReceiver_Allocator;
      typedef core::array<XEventReceiver*,XEventReceiver_Allocator> XEventReceiver_Array;

    public:  // READ & WRITE

      //! Set to true to activate this event receiver.
      //! If false no event will be handled here.
      //! Default=true
      bool _Active;

      //! The identifacation number of this event receiver.
      s32 _Id;

      //! The priority of this event receiver.
      //! This priority is used to control the receiving order
      //! if this event receiver is stored in a list.
      s32 _Priority;

      //! The identification of new user events.
      //! Default=-1
      static s32 _UserData1;
      
    public:

      //! Constructor.
      //! \param priority The priority to control the receiving order.
      //! \param id The identification number of this event receiver.
      XEventReceiver( s32 priority=0, s32 id=-1 )
      :IReferenceCounted()
      ,IEventReceiver()
      ,_Active( true )
      ,_Id( id )
      ,_Priority( priority )
        {}

        //! Destructor.
      virtual ~XEventReceiver()
        { Clear(); }

      //! Clears the list of event receiver.
      void Clear();

      //! Searches for a event receiver.
      //! \param id The identitification number of the event receiver.
      //! \return The address of the event receiver or NULL if not found.
      XEventReceiver *Find( s32 id ) const;

      //! Insert a new event receiver.
      //! Do not change '_Priority' after inserting.
      void Insert( XEventReceiver *p_xer );

      //! Called if an event happened.
      //! \return true if the event was processed.
      virtual bool OnEvent( const SEvent &event );

      //! Remove the given event receiver.
      void Remove( XEventReceiver *p_xer );

      //! Send a new user defined event.
      //! \param code The event parameter 'UserData2'.
      //! 'UserData1' is set to '_UserData1' to differ from old user events.
      //! \return true if the event was handled.
      bool SendUserEvent( s32 code );

      //! Send a new user defined event.
      //! \param codes A list of event parameters 'UserData2'.
      //! See SendUserEvent(s32) too.
      //! \param from The index of the first user event code.
      //! \param to The index of the last user event code.
      void SendUserEvent( s32_Array codes, u32 from=0, u32 to=0xFFFFFFFFL );

    protected:

      //! Called if an GUI event happened.
      //! \return true if the event was processed.
      virtual bool OnGuiEvent( const SEvent::SGUIEvent &event )
        { return false; }

      //! Called if an KEY event happened.
      //! \return true if the event was processed.
      virtual bool OnKeyEvent( const SEvent::SKeyInput &event )
        { return false; }

      //! Called if an LOG event happened.
      //! \return true if the event was processed.
      virtual bool OnLogEvent( const SEvent::SLogEvent &event )
        { return false; }

      //! Called if an MOUSE event happened.
      //! \return true if the event was processed.
      virtual bool OnMouseEvent( const SEvent::SMouseInput &event )
        { return false; }

      //! Called if an old USER event happened.
      //! \return true if the event was processed.
      virtual bool OnUserEvent( const SEvent::SUserEvent &event )
        { return false; }

      //! Called if an new USER event happened.
      //! \return true if the event was processed.
      virtual bool OnUserCodeEvent( s32 code )
        { return false; }

    private:

      // A list of event receiver sorted by priority.
      XEventReceiver_Array _EventReceiver;

  };


  //! Class of a key event receiver as a simple example.
  //! This class will handle all key events.
  class XExampleEventReceiver : public XEventReceiver
  {

    public:  // READ & WRITE

      //! Set true to handle a key.
      //! If set to false this key will be ignored.
      //! Note that handled keys can not be handled by following event receiver.
      //! Default=false (all)
      bool HandleKey[KEY_KEY_CODES_COUNT];

      //! A key pressed flag of every key code.
      //! See "Keycodes.h"
      bool KeyIsDown[KEY_KEY_CODES_COUNT];

      //! Is set to true if a key has been changed.
      //! This flag can be used as a key changed trigger.
      //! Note that you must clear this flag after accessing.
      bool KeyHasBeenChanged[KEY_KEY_CODES_COUNT];
      
    public:
    
      //! Constructor.
      //! \param priority The priority to control the receiving order.
      //! \param id The identification number of this event receiver.
      XExampleEventReceiver( s32 priority=0, s32 id=-1 )
      :XEventReceiver()
        { for ( int i=0; i<KEY_KEY_CODES_COUNT; i++ )
            HandleKey[i] = KeyIsDown[i] = KeyHasBeenChanged[i] = false;
        }

      //! (De-)activates a list of keys.
      //! \param p_keylist A list of keys. Letters are case insensitive.
      //! Note that characters are ignored if there does no key code exist.
      //! \param activate Set to false to deactivate the given keys.
      void installKeys( const char *p_keylist, bool activate=true );

      //! (De-)activates a list of special keys.
      //! \param p_keylist A list of special keys. Special keys are e.g.
      //! digits mapped to the num pad and letters mapped to the F1-F24 keys.
      //! '?' is used for the help key. 
      //! Note that characters are ignored if there does no key code exist.
      //! \param activate Set to false to deactivate the given keys.
      void installSpecialKeys( const char *p_keylist, bool activate=true );

    protected:

      //! Called on a key event. See OnKeyEvent().
      //! \return true if the event was processed.
      virtual bool handleEvent()
        { return false; }

      //! Called if an KEY event happened.
      //! \return true if the event was processed.
      virtual bool OnKeyEvent( const SEvent::SKeyInput &event )
        { if ( !HandleKey[event.Key] )
            return false;
          if ( KeyIsDown[event.Key] != event.PressedDown )
            KeyHasBeenChanged[event.Key] = true;
          KeyIsDown[event.Key] = event.PressedDown;
          handleEvent();
          return true;
        }
        
  };

  
  //! Class of a event receiver translating events into user codes.
  //! Overwrite the virtual method OnUserCodeEvent() to hande the user code.
  class XUserCodeEventReceiver : public XEventReceiver
  {

    public:  // READ & WRITE

      //! The codes of all keys. See getKeyIndex() too.
      s32 _KeyCode[2*2*2*KEY_KEY_CODES_COUNT];

      //! The codes of all mouse events.
      //! See enum EMOUSE_INPUT_EVENT in file "IEventReceiver.h".
      s32 _MouseCode[EMIE_COUNT];

    public:

      //! Constructor.
      //! \param priority The priority to control the receiving order.
      //! \param id The identification number of this event receiver.
      XUserCodeEventReceiver( s32 priority=0, s32 id=-1 )
      :XEventReceiver()
        { Clear(); }

      //! Clears all codes (will be set to zero).
      void Clear()
        { for ( u32 i=0; i<(sizeof(_KeyCode)/sizeof(s32)); i++ )
            _KeyCode[i] = 0;
          for ( u32 i=0; i<(sizeof(_MouseCode)/sizeof(s32)); i++ )
            _MouseCode[i] = 0;
        }
        
      //! Returns the index of array '_KeyCode'.
      //! \param key The key.
      //! \param pressed_down true if the key is pressed down.
      //! \param shift true if shift is also pressed.
      //! \param control true if control is also pressed.
      //! \return The index of array '_KeyCode'.
      u32 getKeyIndex( EKEY_CODE key, bool pressed_down,
                       bool shift, bool control ) const
        { const u32 pdsc = ( ( pressed_down ? 0x01UL : 0x00UL ) |
                             ( shift        ? 0x02UL : 0x00UL ) |
                             ( control      ? 0x04UL : 0x00UL ) );
          return ( pdsc*KEY_KEY_CODES_COUNT + (u32)key );
        }

      //! Set a user code for a key ignoring shift and control.
      //! \param key The key.
      //! \param pressed_down true if the key is pressed down.
      //! \param code The user code.
      void setUserCode( EKEY_CODE key, bool pressed_down, s32 code )
        { _KeyCode[getKeyIndex(key,pressed_down,false,false)] = code;
          _KeyCode[getKeyIndex(key,pressed_down,false, true)] = code;
          _KeyCode[getKeyIndex(key,pressed_down, true,false)] = code;
          _KeyCode[getKeyIndex(key,pressed_down, true, true)] = code;
        }

      //! Set user codes for a key.
      //! \param key The key.
      //! \param pressed_down true if the key is pressed down.
      //! \param code The user code.
      void setUserCodes( EKEY_CODE key, bool pressed_down,
                         s32 code, s32 code_shift, s32 code_control,
                         s32 code_shift_and_control )
        { _KeyCode[getKeyIndex(key,pressed_down,false,false)] = code;
          _KeyCode[getKeyIndex(key,pressed_down,false, true)] = code_control;
          _KeyCode[getKeyIndex(key,pressed_down, true,false)] = code_shift;
          _KeyCode[getKeyIndex(key,pressed_down, true, true)] = code_shift_and_control;
        }

      //! Set a user code for a key ignoring shift.
      //! \param key The key.
      //! \param pressed_down true if the key is pressed down.
      //! \param control true if control is also pressed.
      //! \param code The user code.
      void setUserCodeControl( EKEY_CODE key, bool pressed_down, bool control, s32 code )
        { _KeyCode[getKeyIndex(key,pressed_down,false,control)] = code;
          _KeyCode[getKeyIndex(key,pressed_down, true,control)] = code;
        }

      //! Set a user code for a key ignoring control.
      //! \param key The key.
      //! \param pressed_down true if the key is pressed down.
      //! \param shift true if shift is also pressed.
      //! \param code The user code.
      void setUserCodeShift( EKEY_CODE key, bool pressed_down, bool shift, s32 code )
        { _KeyCode[getKeyIndex(key,pressed_down,shift,false)] = code;
          _KeyCode[getKeyIndex(key,pressed_down,shift, true)] = code;
        }

    protected:

      //! Called if an KEY event happened.
      //! \return true if the event was processed.
      virtual bool OnKeyEvent( const SEvent::SKeyInput &event )
        { return SendUserEvent
                 ( _KeyCode[getKeyIndex(event.Key, event.PressedDown,
                                        event.Shift, event.Control)] );
        }

      //! Called if an MOUSE event happened.
      //! \return true if the event was processed.
      virtual bool OnMouseEvent( const SEvent::SMouseInput &event )
        { return SendUserEvent( _MouseCode[event.Event] ); }

      //! Called if an new USER event happened.
      //! \return true if the event was processed.
      virtual bool OnUserCodeEvent( s32 code )
        { return false; }

  };

  
} // end namespace nth
} // end namespace irr

#endif

