// Copyright (C) 2008 Helmut Hirtes
// This file is a free extension of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#include "XMain.h"
#include <iostream>  


namespace irr
{
namespace nth
{

//! constructor (window mode).
XMain::XMain()
:ExitOnEscape( true )
,_bits( 16 )
,_deviceType( EDT_OPENGL )
,p_eventReceiver( this )
,_fullscreen( false )
,_stencilbuffer( false )
,_vsync( false )
,_windowSize( 640, 480 )
{
  // clear all keys
  for ( int i=0; i<KEY_KEY_CODES_COUNT; i++ )
  {
    KeyIsDown[i] = false;
    KeyHasBeenChanged[i] = false;
  }
}


//! Destructor (destroys the device if open)
XMain::~XMain()
{
  destroy();
}


//! Asks the user for the driver type to be used by create().
//! This will work in a shell window only.
//! \return false if the user want you to exit program.
bool XMain::askForDriverType( const char *p_text )
{
  if ( p_text == NULL )
    p_text = "Please select the driver you want for this example";
  // ask user
  printf( p_text );
  printf( ":\n" );
  printf( " (a) Direct3D 9.0c\n" );
  printf( " (b) Direct3D 8.1\n" );
  printf( " (c) OpenGL 1.5\n" );
  printf( " (d) Software Renderer\n" );
  printf( " (e) Burning's Software Renderer\n" );
  printf( " (f) NullDevice\n" );
  printf( " (otherKey) exit\n\n" );
  // input
  char input;
  std::cin >> input;
  // check input
  video::E_DRIVER_TYPE driverType;
  switch ( input )
  {
    case 'a':
    { driverType = EDT_DIRECT3D9;
    } break;
    case 'b': 
    { driverType = EDT_DIRECT3D8;
    } break;
    case 'c':
    { driverType = EDT_OPENGL;
    } break;
    case 'd':
    { driverType = EDT_SOFTWARE;
    } break;
    case 'e': 
    { driverType = EDT_BURNINGSVIDEO;
    } break;
    case 'f':
    { driverType = EDT_NULL;
    } break;
    default:
    { // exit program!
    } return false;
  }
  // set driver type
  setDeviceType( driverType );
  // end
  return true;
}


//! Creates the irrlicht device.
//! \return true if the device could be created.
bool XMain::create()
{
  // check if already created
  if ( !device )
  {
    // try to create a new device
    device.take() =
      createDevice( _deviceType,
                    _windowSize,
                    _bits,
                    _fullscreen,
                    _stencilbuffer,
                    _vsync,
                    p_eventReceiver );
    // get some pointer
    if ( device )
    {
      driver = device->getVideoDriver();
      guienv = device->getGUIEnvironment();
      smgr   = device->getSceneManager();
    }
    // set window title
    if ( !_fullscreen )
      device->setWindowCaption( _windowTitle.c_str() );
  }
  // end
  return ( device );
}


//! Destroys the device if created.
void XMain::destroy()
{
  // first set all pointer to NULL
  driver = NULL;
  guienv = NULL;
  smgr   = NULL;
  // then destroy the device
  device = NULL;  // note: the pointer template will call drop() automatically
}


//! Draws all.
//! Call this method between driver->beginScene() and driver->endScene().
void XMain::drawAll()
{
  installExit();
  smgr->drawAll();
  guienv->drawAll();
}


//! Called on a key event. See OnKeyEvent().
bool XMain::handleEvent()
{
  if ( isOpen() )
  {
    // handle event
    if ( ExitOnEscape )
    if ( KeyIsDown[KEY_ESCAPE] )
    {
      // close device now
      device->closeDevice();
      return true;
    }
  }
  // end
  return false;
}


//! Shows a short information about the used device and frame per second.
//! Here the window title is used to show information (window mode only).
void XMain::showInfo( const wchar_t *p_text )
{
  if ( isWindow() )
  {
    stringw str;
    if ( p_text != NULL )
      str = p_text;
    else
      str = L"Irrlicht Engine";
    str += L" [";
    str += driver->getName();
    str += L"] FPS: ";
    str += (s32)driver->getFPS();
    device->setWindowCaption( str.c_str() );
  }
}


} // end namespace nth
} // end namespace irr

