// Copyright (C) 2008 Helmut Hirtes
// This file is a free extension of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#ifndef __X_MAIN_H_INCLUDED__
#define __X_MAIN_H_INCLUDED__


#include "XEventReceiver.h"
#include "XNiceToHave.h"


// the irrlicht-lib (allows us to use the irrlich-dll easily)
#ifndef _X_DO_NOT_USE_IRR_LIB
#ifdef  _IRR_WINDOWS_
#pragma comment(lib, "Irrlicht.lib")
#endif
#endif


namespace irr
{
namespace nth
{


using namespace irr;
USE_NON_NTH_NAMESPACE


// define to declare the main function easily
// (using this macro you will never forget parameter types of main function)
// _MAIN .. give the name of your program entry point here.
#define VOID_MAIN(_MAIN) \
        void main() { _MAIN(); }
#define INT_MAIN(_MAIN) \
        int main() { return _MAIN(); }
#define INT_MAIN_ARG(_MAIN) \
        int main( int argc, char* argv[] ) { return _MAIN( argc, argv ); }
#define INT_WIN_MAIN2(_MAIN) \
        INT WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR lpCmdLine, INT ) \
        { return _MAIN( hInstance, lpCmdLine ); }
#define INT_WIN_MAIN4(_MAIN) \
        INT WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, \
                            LPSTR lpCmdLine, INT nCmdShow ) \
        { return _MAIN( hInstance, hPrevInstance, lpCmdLine, nCmdShow ); }



// some useful pointer templates
typedef pointer<ICameraSceneNode>      ICameraSceneNode_ptr;
typedef pointer<IGUIInOutFader>        IGUIInOutFader_ptr;
typedef pointer<IGUIEnvironment>       IGUIEnvironment_ptr;
typedef pointer<IQ3LevelMesh>          IQ3LevelMesh_ptr;
typedef pointer<IrrlichtDevice>        IrrlichtDevice_ptr;
typedef pointer<ISceneManager>         ISceneManager_ptr;
typedef pointer<ISceneNode>            ISceneNode_ptr;
typedef pointer<ISceneNodeAnimator>    ISceneNodeAnimator_ptr;
typedef pointer<ITriangleSelector>     ITriangleSelector_ptr;
typedef pointer<IVideoDriver>          IVideoDriver_ptr;


  //! Class of the main window. The irrlicht device will be created here.
  //! See file "irrlicht.h" function createDevice().
  //! Note: Overwrite OnEvent() to handle your own events.
  class XMain : public XExampleEventReceiver
  {

    public:  // READ & WRITE

      //! Set to true to exit if escape is pressed. Default=true
      //! Note that this will work on the base class (XExampleEventReceiver)
      //! only. All other event receiver will ignore this flag.
      //! To (de)activate this feature do a call to drawAll() or installExit().
      bool ExitOnEscape;

      //! A call to this method will install the exit main event handler.
      //! See 'ExitOnEscape' for more information.
      void installExit()
        { HandleKey[KEY_ESCAPE] = ExitOnEscape; }

    public:  // these pointer are READ ONLY !

      //! The address of the device or NULL.
      IrrlichtDevice_ptr device;

      //! The address of the GUI environment or NULL.
      IGUIEnvironment_ptr guienv;

      //! The address of the scene manager or NULL.
      ISceneManager_ptr smgr;

      //! The address of the video driver or NULL.
      IVideoDriver_ptr driver;

    public:

   		//! Constructor.
      //! Default values:
      //! - depth = 16
      //! - device type = EDT_OPENGL
      //! - this class is used as event receiver
      //! - window mode (no full screen)
      //! - no stencil buffer
      //! - no vertical synchronisation
      //! - size of window: width=640, height=480
      //! - no window title (empty string)
      XMain();

      //! Destructor (destroys the device if open)
      virtual ~XMain();

      //! Asks the user for the driver type to be used by create().
      //! This will work in a shell window only.
      //! \param p_text give a question string or use the default question.
      //! \return false if the user want you to exit program.
      bool askForDriverType( const char *p_text=NULL );

      //! Creates the irrlicht device using the internal stored attribtes.
      //! \return true if the device could be created.
      bool create();

      //! Creates the irrlicht device using the internal stored attribtes.
      //! \param bits set to the depth (16 or 32).
      //! \return true if the device could be created.
      bool createFullScreen( u32 bits )
        { setFullScreen( bits );
          return create();
        }

      //! Creates the irrlicht device using the internal stored attribtes.
      //! \param windowSize set to width and height of window.
      //! \param windowTitle set to the title of window. e.g. L"Title"
      //! \return true if the device could be created.
      bool createWindow( const dimension2d<s32> &windowSize,
                         const wchar_t *windowTitle )
        { setWindow( windowSize, windowTitle );
          return create();
        }

      //! Destroys the device if created.
      void destroy();

      //! Draws all.
      //! Call this method between driver->beginScene() and driver->endScene().
      virtual void drawAll();

      //! Returns the device type.
      E_DRIVER_TYPE getDeviceType() const
        { return _deviceType; }

      //! Returns the event receiver or NULL if not existing.
      IEventReceiver *getEventReceiver()
        { return p_eventReceiver; }

      //! Returns the height of window.
      s32 getHeight() const
        { return _windowSize.Height; }

      //! Returns the width of window.
      s32 getWidth() const
        { return _windowSize.Width; }

      //! Returns the size of window.
      const dimension2d<s32> &getWindowSize() const
        { return _windowSize; }

      //! Returns the title of window.
      const stringw &getWindowTitle() const
        { return _windowTitle; }

      //! Called on a key event. See OnKeyEvent().
      //! \return true if the event was processed.
      virtual bool handleEvent();
      
      //! Checks if full screen mode is active.
      bool isFullScreen()
        { return _fullscreen; }

      //! Checks if the device is open (created).
      bool isOpen()
        { return device.isValid(); }

      //! Checks if the stencil buffer is active.
      bool isStencilBuffer()
        { return _stencilbuffer; }

      //! Checks if the vertical synchronisation is active.
      bool isVSync()
        { return _vsync; }

      //! Checks if window mode is active.
      bool isWindow()
        { return !_fullscreen; }

      //! Runs the driver if existing.
      //! \return true while driver is running and window not closed.
      bool run()
        { if ( isOpen() )
            return device->run();
          return false;  
        }

      //! Shows a short information about the used device and frame per second.
      //! Here the window title is used to show information (window mode only).
      //! Note: Use a timer or a counter (calls of run()) to call this method.
      //! \param p_text will be shown in front of the information.
      void showInfo( const wchar_t *p_text=NULL );

      //! Sets the type of video device.
      //! \param deviceType set to EDT_NULL, EDT_SOFTWARE, EDT_BURNINGSVIDEO,
      //! EDT_DIRECT3D8, EDT_DIRECT3D9 or EDT_OPENGL.
      void setDeviceType( E_DRIVER_TYPE deviceType )
        { _deviceType = deviceType; }

      //! Sets the event receiver.
      //! Note that this will not affect a device that is already open.
      //! \param p_receiver give a reference to your event receiver.
      void setEventReceiver( IEventReceiver *p_receiver=NULL )
        { p_eventReceiver = p_receiver; }

      //! Sets the parameters of the full screen mode.
      //! \param bits set to the depth (16 or 32).
      void setFullScreen( u32 bits )
        { _fullscreen = true;
          _bits       = bits;
        }

      //! (De)actiavtes the stencil buffer.
      //! \param active set to true to activate the stencil buffer.
      void setStencilBuffer( bool active )
        { _stencilbuffer = active; }

      //! (De)actiavtes the vertical synchronisation.
      //! \param active set to true to activate the vertical synchronisation.
      void setVSync( bool active )
        { _vsync = active; }

      //! Sets the parameters of a window mode.
      //! \param windowSize set to width and height of window.
      //! \param windowTitle set to the title of window. e.g. L"Title"
      void setWindow( const dimension2d<s32> &windowSize,
                      const wchar_t *windowTitle )
        { _fullscreen  = false;
          _windowSize  = windowSize;
          _windowTitle = ( (windowTitle==NULL) ? L"" : windowTitle );
        }

    private:

      //! The depth (16 or 32 bits).
      u32 _bits;

      //! The type of device. 
      //! See file "EDriverTypes.h" enum E_DRIVER_TYPE.
      // EDT_NULL          .. Null driver, useful for applications to run the
      //                      engine without visualisation. The null device is
      //                      able to load textures, but does not render and
      //                      display any graphics.
      // EDT_SOFTWARE      .. The Irrlicht Engine Software renderer, runs on
      //                      all platforms, with every hardware. It should
      //                      only be used for 2d graphics, but it can also
      //                      perform some primitive 3d functions. These 3d
      //                      drawing functions are quite fast, but very
      //                      inaccurate, and don't even support clipping in
      //                      3D mode.
      // EDT_BURNINGSVIDEO .. The Burning's Software Renderer, an alternative
      //                      software renderer for Irrlicht. Basically it can
      //                      be described as the Irrlicht Software renderer on
      //                      steroids. It rasterizes 3D geometry perfectly:
      //                      It is able to perform correct 3d clipping,
      //                      perspective correct texture mapping, perspective
      //                      correct color mapping, and renders sub pixel
      //                      correct, sub texel correct primitives. In
      //                      addition, it does bilinear texel filtering and
      //                      supports more materials than the EDT_SOFTWARE
      //                      driver. This renderer has been written entirely
      //                      by Thomas Alten, thanks a lot for this huge
      //                      contribution.
      // EDT_DIRECT3D8     .. Direct3D8 device, only available on Win32
      //                      platforms. Performs hardware accelerated
      //                      rendering of 3D and 2D primitives.
      // EDT_DIRECT3D9     .. Direct3D 9 device, only available on Win32
      //                      platforms. Performs hardware accelerated
      //                      rendering of 3D and 2D primitives.
      // EDT_OPENGL        .. OpenGL device, available on most platforms.
      //                      Performs hardware accelerated rendering of 3D and
      //                      2D primitives.
      video::E_DRIVER_TYPE _deviceType;

      //! The event receiver or NULL.
      IEventReceiver *p_eventReceiver;

      //! Open full screen?
      bool _fullscreen;

      //! Using the stencil buffer?
      bool _stencilbuffer;

      //! Using vertical synchronisation?
      bool _vsync;

      //! The size of window (window mode only).
      dimension2d<s32> _windowSize;

      //! The title of the window (window mode only).
      stringw _windowTitle;

  };

  
} // end namespace nth
} // end namespace irr

#endif

