// Copyright (C) 2008 Helmut Hirtes
// This file is a free extension of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#ifndef __X_NICE_TO_HAVE_H_INCLUDED__
#define __X_NICE_TO_HAVE_H_INCLUDED__


#include <irrlicht.h>


// a macro to use the most important sub namespaces
#define USE_NAMESPACE \
        USE_NON_NTH_NAMESPACE \
        using namespace nth;
#define USE_NON_NTH_NAMESPACE \
        using namespace irr; \
        using namespace core; \
        using namespace scene; \
        using namespace video; \
        using namespace io; \
        using namespace gui;


namespace irr
{
namespace nth
{


  //! This pointer template handles references to classes derived from
  //! base class IReferenceCounted. The method drop() will be called
  //! automatically. Do not care about this. The only thing you have to
  //! do is to use the method take() to take references given by methods
  //! called createXxx().
  template <class cT>
  class pointer
  {

    public:

      //! Constructor. Creates a NULL reference.
      pointer()
      :ptr( NULL )
        {}

      //! Destructor. Drops the reference if valid.
      ~pointer()
        { drop(); }

      //! Drops the reference if valid.
      void drop()
        { if ( isValid() )
          { ptr->drop();
            ptr = NULL;
          }
        }

      //! Checks if the reference is NULL.
      //! /return true if the reference is NULL.
      bool isNull() const
        { return ( ptr == NULL ); }

      //! Checks if the reference is valid.
      //! /return true if the reference is valid.
      bool isValid() const
        { return ( ptr != NULL ); }

      //! Check operator. Usage: if ( pointer ) {...}
      operator bool() const
        { return isValid(); }

      //! Check operator. Usage: if ( !pointer ) {...}
      bool operator !() const
        { return !isValid(); }

      //! Access operator.
      cT *operator->()
        { if ( isNull() )
            throw "NULL pointer access";
          return ptr;
        }

      //! Access operator.
      operator cT*()
        { if ( isNull() )
            throw "NULL pointer access";
          return ptr;
        }

      //! Access operator.
      operator cT&()
        { if ( isNull() )
            throw "NULL pointer access";
          return *ptr;
        }

      //! Assignment operator. Note that grab() will be called here.
      //! Do never use this assignment operator to take the result of
      //! a createXxx() method. Use the method take() instead.
      void operator=( cT *p_class )
        { drop();
          ptr = p_class;
          if ( isValid() )
            ptr->grab();
        }

      //! Assignment operator. Note that grab() will be called here.
      //! Do never use this assignment operator to take the result of
      //! a createXxx() method. Use the method take() instead.
      void operator=( cT &r_class )
        { drop();
          ptr = &r_class;
          if ( isValid() )
            ptr->grab();
        }

      //! Assignment operator. Note that grab() will be called here.
      //! Do never use this assignment operator to take the result of
      //! a createXxx() method. Use the method take() instead.
      void operator=( pointer<cT> p )
        { drop();
          ptr = p.ptr;
          if ( isValid() )
            ptr->grab();
        }

      //! Takes the given reference. Note that grab() will not be called here.
      //! Use this method to store the result of a method called createXxx().
      //! Usage: my_ptr.take() = createXxx();
      cT* &take()
        { drop();
          return ptr;
        }

    private:

      //! The reference address or NULL.
      cT *ptr;

  };


} // end namespace nth
} // end namespace irr

#endif
